from enum import auto, Flag
from functools import reduce
from typing import Optional, Final

from . import Endianness


_SHIFT: Final[int] = 8


class Attribute(Flag):
    STANDOUT   = 1 << 8 + _SHIFT
    UNDERLINE  = auto()
    REVERSE    = auto()
    BLINK      = auto()
    DIM        = auto()
    BOLD       = auto()
    ALTCHARSET = auto()
    INVISIBLE  = auto()
    PROTECT    = auto()
    HORIZONTAL = auto()
    LEFT       = auto()
    LOW        = auto()
    RIGHT      = auto()
    TOP        = auto()
    VERTICAL   = auto()
    NONE       = 0

    @property
    def as_string(self) -> Optional[str]:
        if self == Attribute.NONE:
            return None

        flags = {flag for flag in Attribute.__members__.values() if flag in self and flag is not Attribute.NONE}

        return " + ".join(map(lambda flag: flag.name.capitalize(), flags))

    def get_code(self, length: int = 4, byteorder: Endianness = "little", signed: bool = False) -> bytes:
        return self.value.to_bytes(length=length, byteorder=byteorder, signed=signed)

    @staticmethod
    def from_names(names: Optional[str]) -> "Attribute":
        if names is None:
            return Attribute.NONE

        return reduce(lambda a, b: a | b, map(lambda k: getattr(Attribute, k.upper(), Attribute.NONE), names.replace(' ', '').split('+')))

    @staticmethod
    def from_code(value: bytes, byteorder: Endianness = "little", signed: bool = True) -> "Attribute":
        return Attribute(int.from_bytes(value, signed=signed, byteorder=byteorder))



import mmap
import struct
from functools import cache
from pathlib import Path
from subprocess import run
from typing import Optional, Final

from . import Endianness
from .sequence import Sequence
from .attribute import Attribute
from .color import Color
from .config import Config
from .key import Key
from .theme import Theme


class NcduColors:
    def __init__(self, ncdu: Path):
        self.ncdu: Path = ncdu

        self.version: tuple[int] = self._load_version()
        self.supports_darkbg = self.version >= (1, 17)

        self.binary: bytearray = bytearray(self.ncdu.read_bytes())

        # See https://en.wikipedia.org/wiki/Executable_and_Linkable_Format
        if self.binary[5] == 1:
            self.byteorder: Endianness = "little"
        elif self.binary[5] == 2:
            # self.byteorder: Endianness = "big"
            raise NotImplementedError("Big endian ELF files are not supported yet.")
        else:
            raise ValueError("Malformed ELF file.")

    @cache
    def _load_version(self) -> tuple[int]:
        command = run([self.ncdu, "--version"], capture_output=True)

        command.check_returncode()

        ncdu_literally, raw_version = command.stdout.decode("utf-8").split(maxsplit=1)

        if ncdu_literally != "ncdu":
            raise ValueError(f"Executable {str(self.ncdu.absolute())!r} was not recognised as Ncdu.")

        version = tuple(map(int, raw_version.rsplit('-')[0].split('.')))

        if version >= (2, 0):
            raise ValueError(f"Version 2.0+ ({raw_version}) is not supported yet.")

        return version

    @staticmethod
    def binary_to_themes(*, binary: bytearray, offset: int, supports_darkbg: bool, byteorder: Endianness) -> tuple[Theme]:
        if supports_darkbg:
            themes: tuple[Theme] = Theme("off"), Theme("dark"), Theme("darkbg")
        else:
            themes: tuple[Theme] = Theme("off"), Theme("dark")

        key_format: Final[str] = Key.get_format(byteorder=byteorder)

        expected_length: int = struct.calcsize(key_format)

        full_binary_segment: bytearray = binary[offset: offset + expected_length * len(Theme.KEYS) * len(themes)]

        for i, key_str in enumerate(Theme.KEYS):
            for j, theme in enumerate(themes):
                index = (len(themes) * i + j) * expected_length

                binary_segment: bytearray = full_binary_segment[index : index + expected_length]

                assert len(binary_segment) == expected_length

                fg_raw, bg_raw, a_raw = struct.unpack(key_format, binary_segment)

                key = Key(
                    fg=Color.by_value(fg_raw),
                    bg=Color.by_value(bg_raw),
                    a=Attribute(a_raw)
                )

                setattr(theme, key_str, key)

        return themes

    def load_config(self, offset: int) -> Config:
        themes: tuple[Theme] = NcduColors.binary_to_themes(
            binary=self.binary,
            offset=offset,
            supports_darkbg=self.supports_darkbg,
            byteorder=self.byteorder
        )

        return Config(ncdu=self.ncdu, offset=offset, **{theme.name: theme for theme in themes})

    def extract_default_config(self) -> Config:
        if self.byteorder == "little":  # todo: big endian sequences
            offset = self.binary.find(Sequence.DEFAULT_LE_WITH_DARKBG if self.supports_darkbg else Sequence.DEFAULT_LE_WITHOUT_DARKBG)
        else:  # todo: big endian sequences
            raise NotImplementedError("Big endian ELF files are not supported yet.")

        if offset == -1:
            raise ValueError("Default config pattern not found in the binary file.\n"
                             "You can only to do a 'apply-config' (on the default config) or 'revert'.")

        default_config: Config = self.load_config(offset=offset)

        return default_config

    @staticmethod
    def dump_internal_default_config(*, ncdu: Optional["Path"] = None, offset: Optional[int] = None, with_darkbg: bool,
                                     byteorder: Endianness = "little") -> Config:
        if byteorder == "little":
            binary = Sequence.DEFAULT_LE_WITH_DARKBG if with_darkbg else Sequence.DEFAULT_LE_WITHOUT_DARKBG
        else:  # todo: big endian sequences
            raise NotImplementedError("Big endian ELF files are not supported yet.")

        themes = NcduColors.binary_to_themes(binary=binary, offset=0, supports_darkbg=with_darkbg, byteorder=byteorder)

        return Config(
            ncdu=ncdu,
            offset=offset,
            **{theme.name: theme for theme in themes}
        )

    def apply_config(self, new_config: Config) -> bool:
        offset: int = new_config.offset

        current_config: Config = self.load_config(offset=offset)

        if current_config.as_dict() == new_config.as_dict():
            return False

        with open(self.ncdu, "r+b") as edit:
            new_bytes = new_config.as_bytes()

            self.binary[offset: offset + len(new_bytes)] = new_bytes

            with mmap.mmap(edit.fileno(), 0) as mem:
                mem.seek(offset)
                mem.write(new_bytes)

        return True



import argparse
import json
import shutil
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import Optional

from . import __version__
from ncducolors.config import Config
from ncducolors.ncducolors import NcduColors


HELP_MESSAGE = """
NcduColors dumps and patches pre-2.0 Ncdu's internal themes.

Usage:
  ncducolors [--ncdu PATH] extract-default-config <FILE> [--compact]
  ncducolors [--ncdu PATH] apply-config <FILE>
  ncducolors [--ncdu PATH] revert (--offset INT | --config FILE)
  ncducolors [--ncdu PATH] dump-internal-default-config [--compact]
             [--with[out]-darkbg] [--[little|big]-endian]
  ncducolors (-h | --help)
  ncducolors --version

Actions:
  extract-default-config         Extracts Ncdu's configuration, its two/three themes,
                                 colors and attributes. Works only on non-patched binaries.
                                 (make a backup of it before proceding!)
  apply-config                   Overwrites the binary configuration file with the one
                                 provided.
  revert                         Does the same thing of apply-config, but uses
                                 NcduColors' colors and attributes. In other words,
                                 everything else than "path" and "offset" is discarded.
  dump-internal-default-config   Should be used in exceptional cases only
                                 (last-resort recovery, analysis, etc...).
                                 It uses NcduColors' (not Ncdu's) binaries.

Options:
  -h --help        Show this screen.
  --version        Show version and exit.
  --ncdu PATH      Use the provided Ncdu binary as reference for some default values
                   (default: Ncdu is automatically recognised).
  --big-endian     Force the dumping of the internal default config used by Ncdu on
                   big-endian machines (default: depends on the Ncdu binary).
  --little-endian  Like --big-endian, but for little endian binaries.
  --with-darkbg    Force the dumping of the internal default config used by Ncdu >= 1.7.
  --without-darkbg Force the dumping of the internal default config used by Ncdu < 1.7.
  --compact        Disable the JSON indentation.
  --offset INT     Use the provided offset (it's binary-dependent - be careful).

How to use this software:
1. Use 'extract-default-config' to extract your config to a JSON file.
2. Make a backup of the mentioned JSON file.
3. Use your editor of choice to edit the *values* (also: DON'T edit "offset").
4. Use 'apply-config' to apply the new configuration file.
5. Done. Launch Ncdu. Or re-start from step 3 / 'revert' using step 2's backup.

Config keys and values:
- "ncdu" is the Path to the Ncdu binary and "offset" depends on it. Don't touch them.
- "off", "dark" and (optionally) "darkbg" are the themes' names.
- "fg" and "bg" support the colors: Black, Red, Green, Yellow, Blue, Magenta, Cyan, White.
- "a" supports the attributes: Standout, Underline, Reverse, Blink, Dim, Bold, Altcharset,
  Invisible, Protect, Horizontal, Left, Low, Right, Top, Vertical (concatenate with '+').
- The value "null" means default foreground color, background color or all attributes off.
- Colors and attributes are case-insensitive.

In case something goes very wrong:
  NcduColors does some checks before applying any new configuration file.
  However, something could still go wrong if you use this tool in the wrong way.
  In case something bad happens, please reinstall Ncdu using your package manager.

For more help: see https://gitlab.com/Midblyte/NcduColors.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT.
""".strip()


class Handlers:
    COMPACT_ARGS = {"indent": None, "separators": (",", ":")}
    EXPANDED_ARGS = {"indent": 4}

    @staticmethod
    def dump_internal_default_config(ncdu: NcduColors, compact: bool, config: TextIOWrapper, darkbg: Optional[bool] = None, byteorder: Optional[str] = None):
        internal_default_config = NcduColors.dump_internal_default_config(with_darkbg=darkbg or ncdu.supports_darkbg, byteorder=byteorder or ncdu.byteorder)

        kwargs = Handlers.COMPACT_ARGS if compact else Handlers.EXPANDED_ARGS

        isatty = config.isatty()

        with config as file:
            file.write(json.dumps(internal_default_config.as_dict(), **kwargs))

        if isatty:
            print("Config dumped successfully.")

    @staticmethod
    def extract_default_config(ncdu: NcduColors, compact: bool, config: TextIOWrapper):
        extracted_default_config = ncdu.extract_default_config()

        kwargs = Handlers.COMPACT_ARGS if compact else Handlers.EXPANDED_ARGS

        isatty = config.isatty()

        with config as file:
            file.write(json.dumps(extracted_default_config.as_dict(), **kwargs))

        if isatty:
            print("Config extracted successfully.")

    @staticmethod
    def apply_config(ncdu: NcduColors, config: TextIOWrapper):
        new_config = Config.from_buffer(config)

        if ncdu.apply_config(new_config=new_config):
            print("Config applied successfully.")
        else:
            print("Config is already applied.")

    @staticmethod
    def revert(ncdu: NcduColors, config: Optional[TextIOWrapper] = None, offset: Optional[int] = None):
        internal_default_config: Config = NcduColors.dump_internal_default_config(
            ncdu=ncdu.ncdu,
            offset=offset or Config.from_buffer(config).offset,
            with_darkbg=ncdu.supports_darkbg,
            byteorder=ncdu.byteorder
        )

        if ncdu.apply_config(new_config=internal_default_config):
            print("Ncdu defaults reverted successfully.")
        else:
            print("Ncdu has already been reverted to its defaults.")


def get_parser():
    parser = argparse.ArgumentParser(prog="ncducolors", usage="%(prog)s [--ncdu PATH] <action> [...]", add_help=False)
    parser.add_argument("--help", action="store_true", default=argparse.SUPPRESS, help="Show this help message and exit")
    parser.add_argument("--ncdu", type=Path, default=shutil.which("ncdu"), help="Path of Ncdu binary")
    parser.add_argument("--version", "-v", action="version", version=f"%(prog)s {__version__}")

    subparser = parser.add_subparsers(title="action")

    dump_internal_default_config = subparser.add_parser(
        name="dump-internal-default-config",
        help="Dump this software's internal config. Usually, you shouldn't rely on this config"
    )
    dump_internal_default_config.set_defaults(handler=Handlers.dump_internal_default_config)
    dump_internal_default_config.add_argument("config", type=argparse.FileType("wt"), default=sys.stdout, help="Path of config file (JSON) where to save")
    dump_internal_default_config.add_argument("--compact", action="store_true", help="Don't indent the JSON file")
    dump_internal_default_config_type = dump_internal_default_config.add_mutually_exclusive_group(required=False)
    dump_internal_default_config_type.add_argument("--with-darkbg", dest="darkbg", action="store_true")
    dump_internal_default_config_type.add_argument("--without-darkbg", dest="darkbg", action="store_false")
    dump_internal_default_config_byteorder = dump_internal_default_config.add_mutually_exclusive_group(required=False)
    dump_internal_default_config_byteorder.add_argument("--little-endian", dest="byteorder", action="store_const", const="little")
    dump_internal_default_config_byteorder.add_argument("--big-endian", dest="byteorder", action="store_const", const="big")

    extract_default_config = subparser.add_parser(name="extract-default-config", help="Extract the original binary config converted to JSON")
    extract_default_config.set_defaults(handler=Handlers.extract_default_config)
    extract_default_config.add_argument("config", type=argparse.FileType("wt"), default=sys.stdout, help="Path of config file (JSON) where to extract")
    extract_default_config.add_argument("--compact", action="store_true", help="Don't indent the JSON file")

    apply_config = subparser.add_parser(name="apply-config", help="Apply an edited config")
    apply_config.set_defaults(handler=Handlers.apply_config)
    apply_config.add_argument("config", type=argparse.FileType("rt"), help="Path of config file (JSON) to apply")

    revert = subparser.add_parser(name="revert", help="Revert Ncdu as it was before being patched")
    revert.set_defaults(handler=Handlers.revert)
    revert_args = revert.add_mutually_exclusive_group(required=True)
    revert_args.add_argument("--config", type=argparse.FileType("rt"), help="Path of config file to load the offset from")
    revert_args.add_argument("--offset", type=int, help="Offset to Ncdu binary's colors config")

    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()

    try:
        if args.ncdu is None:
            raise ValueError("Ncdu was not found.")

        if hasattr(args, "help") or not hasattr(args, "handler"):
            return print(HELP_MESSAGE)

        if (handler := getattr(args, "handler", None)) is not None:
            ncdu = NcduColors(ncdu=args.ncdu.expanduser())

            delattr(args, "ncdu")
            delattr(args, "handler")

            handler(ncdu=ncdu, **args.__dict__)
        else:
            parser.print_help()
    except Exception as exception:
        print(f"Error: {exception}", end="\n\n")

        parser.print_usage()

        exit(1)


if __name__ == "__main__":
    main()



from enum import Enum, auto
from typing import Optional

from . import Endianness


class Color(Enum):
    BLACK   = 0
    RED     = auto()
    GREEN   = auto()
    YELLOW  = auto()
    BLUE    = auto()
    MAGENTA = auto()
    CYAN    = auto()
    WHITE   = auto()
    NONE    = -1
    UNKNOWN = NONE

    @property
    def as_string(self) -> Optional[str]:
        return self.name.capitalize() if self is not Color.NONE else None

    def get_code(self, length: int = 2, byteorder: Endianness = "little", signed: bool = True) -> bytes:
        return self.value.to_bytes(length=length, byteorder=byteorder, signed=signed)

    @staticmethod
    def from_name(name: Optional[str]) -> "Color":
        return getattr(Color, name.upper(), Color.UNKNOWN) if name is not None else Color.NONE

    @staticmethod
    def from_code(value: bytes, byteorder: Endianness = "little", signed: bool = True) -> "Color":
        return Color.by_value(int.from_bytes(value, signed=signed, byteorder=byteorder))

    @classmethod
    def by_value(cls, value: int):
        for member in cls:
            if member.value == value:
                return member

        return Color.UNKNOWN



from functools import cache

from ncducolors.attribute import Attribute
from ncducolors.color import Color
from ncducolors import Endianness


class Key:
    __slots__ = "fg", "bg", "a"

    def __init__(self, fg: Color, bg: Color, a: Attribute):
        self.fg = fg
        self.bg = bg
        self.a = a

        if self.fg is None:
            y = 2

    @staticmethod
    @cache
    def get_format(byteorder: Endianness = "little") -> str:
        return f"{'<' if byteorder == 'little' else '>'}hhI"

    @staticmethod
    def from_dict(dct: dict) -> "Key":
        if set(Key.__slots__) != set(dct.keys()):
            raise ValueError(f"Key object must have keys 'fg', 'bg' and 'a'.")

        fg = Color.from_name(dct["fg"])
        bg = Color.from_name(dct["bg"])
        a = Attribute.from_names(dct["a"])

        key = Key(fg=fg, bg=bg, a=a)

        return key

    def as_dict(self):
        return {
            "fg": self.fg.as_string,
            "bg": self.bg.as_string,
            "a": self.a.as_string
        }

    def as_bytes(self, byteorder: Endianness = "little", **kwargs) -> bytes:
        return b''.join(map(lambda f: f.get_code(byteorder=byteorder, **kwargs), (self.fg, self.bg, self.a)))



import json
from io import TextIOWrapper
from pathlib import Path
from typing import Optional

from ncducolors.theme import Theme


class Config:
    __slots__ = "ncdu", "offset", "off", "dark", "darkbg"

    def __init__(self, ncdu: Path, offset: int, off: Theme, dark: Theme, darkbg: Optional[Theme] = None):
        self.ncdu = ncdu
        self.offset = offset
        self.off = off
        self.dark = dark
        self.darkbg = darkbg

    @staticmethod
    def from_buffer(reader: TextIOWrapper) -> "Config":
        with reader as file:
            config = Config.from_dict(json.loads(file.read()))

        return config

    @staticmethod
    def from_dict(dct: dict) -> "Config":
        if (dct_keys := set(dct.keys())) != (slots := set(Config.__slots__)) and dct_keys != slots - {"darkbg"}:
            raise TypeError("Not a serialized Config object.")

        return Config(
            ncdu=Path(dct["ncdu"]).expanduser(),
            offset=dct["offset"],
            **{k: Theme.from_dict(k, dct[k]) for k in ("off", "dark", "darkbg") if k in dct}
        )

    def as_dict(self) -> dict:
        return {
            "ncdu": str(self.ncdu) if isinstance(self.ncdu, Path) else self.ncdu,
            "offset": self.offset,
            **{k: getattr(self, k).as_dict() for k in (("off", "dark", "darkbg") if self.darkbg is not None else ("off", "dark"))}
        }

    def as_bytes(self) -> bytes:
        if self.darkbg is None:
            themes = self.off, self.dark
        else:
            themes = self.off, self.dark, self.darkbg

        bytes_block = b""

        for str_key in Theme.KEYS:
            for theme in themes:
                bytes_block += getattr(theme, str_key).as_bytes()

        return bytes_block



#!/usr/bin/env python3
import re
from pathlib import Path

from setuptools import setup


with open("README.md", "r") as readme:
    long_description = readme.read()


# Get the version from ncducolors/__init__.py without actually importing the package
__version__ = eval(re.search(r'__version__.+(".+")$', (Path(__file__).with_name("ncducolors") / "__init__.py").read_text(), re.M).group(1))


setup(
    name="NcduColors",
    version=__version__,
    description="Ncdu themes patcher",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Midblyte/NcduColors",
    author="Midblyte",
    author_email="ncducolors@midblyte.anonaddy.me",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Utilities"
    ],
    keywords="ncdu themes",
    project_urls={
        "Source": "https://github.com/Midblyte/NcduColors",
        "Documentation": "https://github.com/Midblyte/NcduColors#how-to-use",
        "Tracker": "https://github.com/Midblyte/NcduColors/issues",
    },
    python_requires=">=3.8",
    packages=["ncducolors"],
    zip_safe=False,
    entry_points={"console_scripts": ["ncducolors = ncducolors.__main__:main"]}
)



from ncducolors.key import Key


class Theme:
    KEYS = (
        "default",
        "box_title",
        "hd",
        "sel",
        "num",
        "num_hd",
        "num_sel",
        "key",
        "key_hd",
        "dir",
        "dir_sel",
        "flag",
        "flag_sel",
        "graph",
        "graph_sel"
    )

    __slots__ = (
        "name",
        *KEYS
    )

    def __init__(self, name: str):
        self.name: str = name

        for key in Theme.KEYS:
            setattr(self, key, ...)

    @staticmethod
    def from_dict(name: str, dct: dict) -> "Theme":
        theme = Theme(name=name)

        for key, value in dct.items():
            if (key := key.lower()) not in Theme.KEYS:
                raise ValueError(f"Key {key!r} is not a valid key for Ncdu themes.")

            setattr(theme, key, Key.from_dict(dct[key]))

        return theme

    def as_dict(self):
        return {
            k: getattr(self, k).as_dict() for k in self.KEYS
        }
